(function() {
  var Client, Message, exports;
  var __slice = Array.prototype.slice, __bind = function(fn, me){ return function(){ return fn.apply(me, arguments); }; };
  exports = this;
  exports.EventEmitter = function() {
    return {
      _events: {},
      on: function(event_name, callback) {
        var _base;
        return ((_base = this._events)[event_name] || (_base[event_name] = [])).push(callback);
      },
      emit: function() {
        var args, event_name, handler, _i, _len, _ref, _results;
        event_name = arguments[0], args = 2 <= arguments.length ? __slice.call(arguments, 1) : [];
        if (this._events[event_name]) {
          _ref = this._events[event_name];
          _results = [];
          for (_i = 0, _len = _ref.length; _i < _len; _i++) {
            handler = _ref[_i];
            _results.push(handler.apply(this, args));
          }
          return _results;
        }
      }
    };
  };
  exports = this;
  exports.Lerpable = function() {
    return {
      _received_states: [],
      lerp_update: function(render_time) {
        var distance_too_great, distance_xy, dont_interpolate, from_state, interpolation_percentage, max_interpolation_distance, offset_time, prop, self, time_between_states, to_state, val, _ref;
        _ref = this.get_states_for_lerping(render_time), from_state = _ref[0], to_state = _ref[1];
        self = this;
        for (prop in from_state) {
          val = from_state[prop];
          self[prop] = val;
        }
        if (!((from_state != null) && (to_state != null))) {
          return;
        }
        max_interpolation_distance = 300;
        distance_xy = [Math.abs(to_state.x - from_state.x), Math.abs(to_state.y - from_state.y)];
        if (distance_xy[0] > max_interpolation_distance || distance_xy[1] > max_interpolation_distance) {
          distance_too_great = true;
        }
        if (distance_too_great) {
          dont_interpolate = true;
        }
        if (dont_interpolate) {
          return interpolation_percentage = 1.0;
        } else {
          offset_time = render_time - from_state.gametime;
          time_between_states = to_state.gametime - from_state.gametime;
          interpolation_percentage = (offset_time / time_between_states) * 1.0;
          if (interpolation_percentage < 0.0) {
            interpolation_percentage = 0;
          }
          if (interpolation_percentage > 1.0) {
            interpolation_percentage = 1;
          }
          this.x += Math.round((to_state.x - from_state.x) * interpolation_percentage);
          return this.y += Math.round((to_state.y - from_state.y) * interpolation_percentage);
        }
      },
      get_states_for_lerping: function(render_time) {
        var from_state, index, state, to_state, _len, _ref;
        from_state = null;
        to_state = null;
        _ref = this._received_states;
        for (index = 0, _len = _ref.length; index < _len; index++) {
          state = _ref[index];
          if (state.gametime > render_time) {
            to_state = state;
            from_state = this._received_states[index - 1];
            break;
          }
        }
        if (to_state && !from_state) {
          from_state = to_state;
        }
        return [from_state, to_state];
      }
    };
  };
  exports = this;
  exports.Message = Message = (function() {
    Message.MESSAGE_TYPES = {
      NEW_PLAYER: 1,
      ACK: 2,
      UPDATE_ENTITY: 3,
      PLAYER_COMMAND: 4,
      GAME_INIT: 5,
      UPDATE_ENTITIES: 6,
      REMOVE_ENTITY: 7
    };
    function Message(attrs) {
      var prop, val;
      for (prop in attrs) {
        val = attrs[prop];
        this[prop] = val;
      }
    }
    return Message;
  })();
  exports = this;
  exports.Client = Client = (function() {
    function Client(args) {
      this.args = args;
      this.LERP = 100;
      this.update_messages_by_entity_id = {};
      this.update_entities_messages = [];
      this.queued_messages = [];
      this.entities_by_id = {};
      this.player_entity_id = null;
      this.next_player_command_id = 0;
      this.unacknowledged_player_commands = [];
      this.last_acked_player_state = null;
      this.socket = io.connect('http://localhost', {
        'force new connection': true,
        reconnect: false
      });
      this.socket.on('connect', function() {
        return console.log('connected');
      });
      this.socket.on('message', __bind(function(data) {
        var msg;
        msg = JSON.parse(data);
        return this.queued_messages.push(msg);
      }, this));
      this.socket.on('disconnect', function() {
        return console.log('disconnected');
      });
    }
    Client.prototype.new_frame = function() {
      var entity, id, messages_to_handle, msg, player, prop, val, _i, _j, _len, _len2, _ref, _ref2, _ref3, _ref4;
      _ref = [this.queued_messages, []], messages_to_handle = _ref[0], this.queued_messages = _ref[1];
      for (_i = 0, _len = messages_to_handle.length; _i < _len; _i++) {
        msg = messages_to_handle[_i];
        this.handle_message(msg);
      }
      player = this.player_entity;
      _ref2 = this.last_acked_player_state;
      for (prop in _ref2) {
        val = _ref2[prop];
        player[prop] = val;
      }
      _ref3 = this.unacknowledged_player_commands;
      for (_j = 0, _len2 = _ref3.length; _j < _len2; _j++) {
        msg = _ref3[_j];
        player[msg.command](msg.args);
      }
      _ref4 = this.entities_by_id;
      for (id in _ref4) {
        entity = _ref4[id];
        if (id !== this.player_entity_id) {
          entity.lerp_update(this.render_time);
        }
      }
      return this.render_time = new Date().getTime() - this.gametime_offset - this.LERP;
    };
    Client.prototype.handle_message = function(message) {
      var entity, prop, val, _ref;
      switch (message.message_type) {
        case Percept.Message.MESSAGE_TYPES.ACK:
          return this.on_ack(message);
        case Percept.Message.MESSAGE_TYPES.GAME_INIT:
          this.player_entity_id = message.your_entity_id;
          entity = this.args.create_entity(message.your_entity_id, message.your_entity_state);
          entity.id = message.your_entity_id;
          this.entities_by_id[entity.id] = entity;
          this.player_entity = entity;
          console.log("my ent id " + entity.id);
          _ref = message.your_entity_state;
          for (prop in _ref) {
            val = _ref[prop];
            entity[prop] = val;
          }
          this.last_acked_player_state = message.your_entity_state;
          this.gametime_offset = new Date().getTime() - message.at_gametime;
          return this.render_time = message.at_gametime - this.LERP - this.FAKE_LAG;
        case Percept.Message.MESSAGE_TYPES.UPDATE_ENTITIES:
          return this.update_entities_states(message);
        case Percept.Message.REMOVE_ENTITY:
          return this.remove_entity(message.entity_id);
      }
    };
    Client.prototype.send_message = function(msg) {
      return this.socket.send(JSON.stringify(msg));
    };
    Client.prototype.update_entities_states = function(msg) {
      var entity, id, state, _ref, _results;
      _ref = msg.entity_states_by_entity_id;
      _results = [];
      for (id in _ref) {
        state = _ref[id];
        entity = this.entities_by_id[id];
        if (!entity) {
          entity = this.args.create_entity(id, state);
          this.entities_by_id[id] = entity;
        }
        state.gametime = msg.at_gametime;
        _results.push(entity._received_states.push(state));
      }
      return _results;
    };
    Client.prototype.remove_entity = function(id) {
      var entity;
      entity = this.entities_by_id[id];
      delete this.entities_by_id[id];
      return this.args.destroy_entity(entity);
    };
    Client.prototype.on_ack = function(message) {
      var cmd;
      if (message.command_id) {
        this.unacknowledged_player_commands = (function() {
          var _i, _len, _ref, _results;
          _ref = this.unacknowledged_player_commands;
          _results = [];
          for (_i = 0, _len = _ref.length; _i < _len; _i++) {
            cmd = _ref[_i];
            if (cmd.command_id > message.command_id) {
              _results.push(cmd);
            }
          }
          return _results;
        }).call(this);
        return this.last_acked_player_state = message.entity_state;
      }
    };
    return Client;
  })();
  exports = this;
  exports.Percept = {
    EventEmitter: EventEmitter,
    Lerpable: Lerpable,
    Message: Message,
    Client: Client
  };
}).call(this);
