window.requestAnimationFrame = (->
  window.requestAnimationFrame or window.webkitRequestAnimationFrame or window.mozRequestAnimationFrame or window.oRequestAnimationFrame or window.msRequestAnimationFrame or ((callback, element) -> window.setTimeout callback, 1000 / 60)
)()


window.client =
  init: ->
    create_entity = (id, state) ->
      # Make a new instance of the MyGameEntity class that we share between 
      # client and server, then add in our client-side dom-specific
      # rendering stuff (an image, so we have something to move around when we get updates)
      e = new window.MyGameEntity()

      # Add in an image attr which we'll use to render
      e.img = '/images/character_boy.png'

      # You MUST give each entity an instance of the Percept.Lerpable properties
      _.extend(e, new Percept.Lerpable())

      # Make this in the DOM
      e.dom_node = $("<img src='#{e.img}'>").css({position: 'absolute'; top: "#{e.x}px", left: "#{e.y}px"}).appendTo('body')

      return e

    destroy_entity = (entity) -> $(entity.dom_node).remove()

    # Here's where the network magic happens:
    # Whenever the server-side game ticks for an update, it sends deltas down the wire to each client.
    # The Percept Client calls lerp_update() on each entity that changed.
    @percept_client = new Percept.Client
      # You MUST provide callbacks to your client code to create and destroy entities.
      create_entity: create_entity 
      destroy_entity: destroy_entity

    @input = new window.InputManager()

  poll_input: ->
    # Update our input manager
    @input.update()

    # Send player movement commands on keyboard input
    movement_keys = {
      up:     87 #w
      down:   83 #s
      left:   65 #a
      right:  68 #d
    }

    dir = null
    dir = 'up'    if @input.isKeyDown(movement_keys.up)
    dir = 'down'  if @input.isKeyDown(movement_keys.down)
    dir = 'left'  if @input.isKeyDown(movement_keys.left)
    dir = 'right' if @input.isKeyDown(movement_keys.right)

    if dir?
      [command_id, @percept_client.next_player_command_id] = [@percept_client.next_player_command_id, @percept_client.next_player_command_id + 1]
      msg = new Percept.Message
        message_type: Percept.Message.MESSAGE_TYPES.PLAYER_COMMAND
        command: 'move'
        command_id: command_id
        args: { dir: dir }

      # Store this command locally for replays until the server acknowledges handling it
      @percept_client.unacknowledged_player_commands.push msg

      # And, of course, send the command to the server
      @percept_client.send_message(msg)

  redraw_entities: ->
    ($(e.dom_node).css(top: e.y, left: e.x)) for id, e of @percept_client.entities_by_id


window.render_frame = (time) ->
  window.requestAnimationFrame render_frame
  window.client.poll_input()
  window.client.percept_client.new_frame()
  window.client.redraw_entities()

$ -> 
  window.client.init()
  window.render_frame()
