#= require './message'

exports = this

exports.Client = class Client
  constructor: (@args) ->
    @LERP = 100
    @update_messages_by_entity_id = {}
    @update_entities_messages = []
    @queued_messages = []
    @entities_by_id = {}
    @player_entity_id = null
    @next_player_command_id = 0
    @unacknowledged_player_commands = []
    @last_acked_player_state = null
    @socket = io.connect('http://localhost', 'force new connection': true, reconnect: false)
    @socket.on 'connect', () ->
      console.log 'connected'
    @socket.on 'message', (data) =>
      # Queue this message for handling on next render frame
      msg = JSON.parse data
      @queued_messages.push msg
    @socket.on 'disconnect', () ->
      console.log 'disconnected'


  new_frame: ->
    # Handle all queued messages (some of these are probably entity updates which we'll interpolate below)
    [messages_to_handle, @queued_messages] = [@queued_messages, []]
    @handle_message(msg) for msg in messages_to_handle

    # Replay un-acknowledged player commands on top of last rendered state for this player (this is Client Input Prediction)
    # 1. First, rewind us back to last acked state
    player = @player_entity
    player[prop] = val for prop, val of @last_acked_player_state
    # 2. Now, replay all stored player commands on the player obj
    player[msg.command](msg.args) for msg in @unacknowledged_player_commands

    # Tell all other changed entities to update themselves.
    for id, entity of @entities_by_id when id != @player_entity_id
      entity.lerp_update(@render_time)

    #Update render time with time_diff since last frame render
    @render_time = new Date().getTime() - @gametime_offset - @LERP


  handle_message: (message) ->
    switch message.message_type
      when Percept.Message.MESSAGE_TYPES.ACK
        @on_ack(message)

      when Percept.Message.MESSAGE_TYPES.GAME_INIT
        @player_entity_id = message.your_entity_id
        entity = @args.create_entity message.your_entity_id, message.your_entity_state
        entity.id = message.your_entity_id
        @entities_by_id[entity.id] = entity
        @player_entity = entity
        console.log "my ent id #{entity.id}"
        entity[prop] = val for prop, val of message.your_entity_state #TODO remove this code duplication of non lerp updating state
        @last_acked_player_state = message.your_entity_state

        # Start our render_time clock at this message's gametime - lerp
        # calculate our offset from servertime so we can turn client time into server time for rendering
        @gametime_offset = new Date().getTime() - message.at_gametime
        @render_time = message.at_gametime - @LERP - @FAKE_LAG

      when Percept.Message.MESSAGE_TYPES.UPDATE_ENTITIES
        @update_entities_states(message)
        # You might be tempted to do:
        # @render_time = message.at_gametime - @LERP 
        # Because, you think you'd want to always update our render time to be @lerp time behind last gamestate message
        # But, changing render time here basically hoses the lerping update algorithm, so for now we'll just trust
        # the client to stay up to date with the server. Maybe later we'll send out periodic SYNC_CLOCK messages...

      when Percept.Message.REMOVE_ENTITY
        @remove_entity(message.entity_id)


  send_message: (msg) ->
    @socket.send JSON.stringify(msg)


  update_entities_states: (msg) ->
    # Pass each new state on to its entity object so it can handle updating itself at the appropriate render_times
    for id, state of msg.entity_states_by_entity_id
      entity = @entities_by_id[id]

      # Create if we don't have this entity yet
      unless entity
        entity = @args.create_entity(id, state) 
        @entities_by_id[id] = entity
      state.gametime = msg.at_gametime
      entity._received_states.push(state) 


  remove_entity: (id) ->
    entity = @entities_by_id[id]
    delete @entities_by_id[id]
    @args.destroy_entity entity # Tell the client about the destroy so it can remove it.


  on_ack: (message) ->
    if message.command_id
      # Remove this player command from our temporary buffer so we don't replay it anymore
      @unacknowledged_player_commands = (cmd for cmd in @unacknowledged_player_commands when cmd.command_id > message.command_id)
      # Record our last acked state
      @last_acked_player_state = message.entity_state


